package de.hft.stuttgart.citydoctor2.healing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.NonManifoldEdgeError;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.Segment3d;
import de.hft.stuttgart.citydoctor2.math.Vector2d;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.utils.Pair;

public class HealNonManifoldEdgeError implements HealingMethod {
	
	private static final Logger logger = LogManager.getLogger(HealNonManifoldEdgeError.class);
	
	@Override
	public boolean visit(NonManifoldEdgeError err, ModificationListener l) {
		logger.debug("Executing Repair for NonManifoldEdgeError");
			// count the number of manifold edges per polygon
			Map<Polygon, AtomicInteger> errorCountMap = new HashMap<>();
			for (Edge e : err.getEdges()) {
				for (Polygon p : e.getAdjacentPolygons()) {
					AtomicInteger count = errorCountMap.computeIfAbsent(p, k -> new AtomicInteger(0));
					count.incrementAndGet();
				}
			}

			// find the polygons with the highest number of error edges
			List<Pair<Polygon, AtomicInteger>> faultyPolygons = new ArrayList<>();
			for (Entry<Polygon, AtomicInteger> e : errorCountMap.entrySet()) {
				faultyPolygons.add(new Pair<Polygon, AtomicInteger>(e.getKey(), e.getValue()));
			}
			Collections.sort(faultyPolygons, (o1, o2) -> o2.getValue1().intValue() - o1.getValue1().intValue());

			int highestValue = faultyPolygons.get(0).getValue1().intValue();
			// remove the polygon with the highest amount, if it is the only one with the
			// highest number
			return removeHighestPolygon(err, errorCountMap, faultyPolygons, highestValue, l);
	}
	
	private boolean removePolygonWithWrongAngle(NonManifoldEdgeError err, ModificationListener l) {
		Edge errorEdge = err.getEdges().get(0);
		List<Polygon> polygons = errorEdge.getAdjacentPolygons();

		Segment3d edgeSegment = new Segment3d(errorEdge.getFrom(), errorEdge.getTo());
		List<Pair<Vector3d, Polygon>> normals = collectNormalsofAdjacentPolygons(errorEdge, polygons, edgeSegment);
		Vector3d center = err.getGeometry().getCenter();
		Vector3d centerNormal = edgeSegment.getNormalThroughPoint(center);
		
		Vector3d edgeDir = errorEdge.getTo().minus(errorEdge.getFrom());
		double x = Math.abs(edgeDir.getX());
		double y = Math.abs(edgeDir.getY());
		double z = Math.abs(edgeDir.getZ());
		
		Vector2d projectedCenterNormal;
		List<Pair<Vector2d, Polygon>> projectedNormals = new ArrayList<>();
		if (x > y && x > z) {
			for (Pair<Vector3d, Polygon> normal : normals) {
				Vector3d normalVec = normal.getValue0();
				Vector2d projectedNormal = new Vector2d(normalVec.getY(), normalVec.getZ());
				projectedNormal.normalize();
				projectedNormals.add(new Pair<>(projectedNormal, normal.getValue1()));
			}
			projectedCenterNormal = new Vector2d(centerNormal.getY(), centerNormal.getZ());
		} else if (y > x && y > z) {
			for (Pair<Vector3d, Polygon> normal : normals) {
				Vector3d normalVec = normal.getValue0();
				Vector2d projectedNormal = new Vector2d(normalVec.getX(), normalVec.getZ());
				projectedNormal.normalize();
				projectedNormals.add(new Pair<>(projectedNormal, normal.getValue1()));
			}
			projectedCenterNormal = new Vector2d(centerNormal.getX(), centerNormal.getZ());
		} else {
			for (Pair<Vector3d, Polygon> normal : normals) {
				Vector3d normalVec = normal.getValue0();
				Vector2d projectedNormal = new Vector2d(normalVec.getX(), normalVec.getY());
				projectedNormal.normalize();
				projectedNormals.add(new Pair<>(projectedNormal, normal.getValue1()));
			}
			projectedCenterNormal = new Vector2d(centerNormal.getX(), centerNormal.getY());
		}
		projectedCenterNormal.normalize();
		double normalAngle = Math.atan2(projectedCenterNormal.getY(), projectedCenterNormal.getX());
		
		removePolygonsWithWrongAngle(err, errorEdge, projectedNormals, normalAngle, l);
		return true;
	}

	private void removePolygonsWithWrongAngle(NonManifoldEdgeError err, Edge errorEdge,
			List<Pair<Vector2d, Polygon>> projectedNormals, double normalAngle, ModificationListener l) {
		Polygon lowestPolygon = null;
		Polygon highestPolygon = null;
		double lowestAngle = 500d;
		double highestAngle = -500d;
		for (Pair<Vector2d, Polygon> normal : projectedNormals) {
			Vector2d normalVec = normal.getValue0();
			double angle = Math.atan2(normalVec.getY(), normalVec.getX());
			angle = angle - normalAngle;
			if (angle > Math.PI) {
				angle -= 2 * Math.PI;
			}
			if (angle < Math.PI) {
				angle += 2 * Math.PI;
			}
			if (angle < lowestAngle) {
				lowestAngle = angle;
				lowestPolygon = normal.getValue1();
			}
			if (angle > highestAngle) {
				highestAngle = angle;
				highestPolygon = normal.getValue1();
			}
		}
		
		for (Polygon p : errorEdge.getAdjacentPolygons()) {
			if (p != lowestPolygon && p != highestPolygon) {
				l.polygonRemoved(p);
				p.remove();
				return;
			}
		}
	}

	private List<Pair<Vector3d, Polygon>> collectNormalsofAdjacentPolygons(Edge errorEdge, List<Polygon> polygons,
			Segment3d edgeSegment) {
		List<Pair<Vector3d, Polygon>> normals = new ArrayList<>();
		for (Polygon p : polygons) {
			for (Vertex v : p.getExteriorRing().getVertices()) {
				if (v != errorEdge.getFrom() && v != errorEdge.getTo()) {
					Vector3d normal = edgeSegment.getNormalThroughPoint(v);
					normals.add(new Pair<>(normal, p));
					break;
				}
			}
		}
		return normals;
	}

	private boolean removeHighestPolygon(NonManifoldEdgeError err, Map<Polygon, AtomicInteger> errorCountMap,
			List<Pair<Polygon, AtomicInteger>> faultyPolygons, int highestValue, ModificationListener l) {
		if (faultyPolygons.get(1).getValue1().intValue() != highestValue) {
			// remove faulty polygon
			Polygon poly = faultyPolygons.get(0).getValue0();
			l.polygonRemoved(poly);
			poly.remove();
			return true;
		} else {
			// two polygons have the same amount of error edges
			// search for more error edges, where number of half edges is < 2 (not closed
			// error)
			searchForNotClosedEdges(err, errorCountMap, faultyPolygons, highestValue);
			
			// resort the polygons again
			Collections.sort(faultyPolygons, (o1, o2) -> o2.getValue1().intValue() - o1.getValue1().intValue());
			
			highestValue = faultyPolygons.get(0).getValue1().intValue();
			if (faultyPolygons.get(1).getValue1().intValue() != highestValue) {
				// remove faulty polygon
				Polygon poly = faultyPolygons.get(0).getValue0();
				l.polygonRemoved(poly);
				poly.remove();
				return true;
			} else {
				removePolygonWithWrongAngle(err, l);
				return true;
			}
		}
	}
	
	private void searchForNotClosedEdges(NonManifoldEdgeError err, Map<Polygon, AtomicInteger> errorCountMap,
			List<Pair<Polygon, AtomicInteger>> faultyPolygons, int highestValue) {
		int index = 0;
		while (index < faultyPolygons.size() && faultyPolygons.get(index).getValue1().intValue() == highestValue) {
			Polygon p = faultyPolygons.get(index).getValue0();
			List<Edge> edges = err.getGeometry().getEdgesAdjacentTo(p);
			for (Edge e : edges) {
				if (e.getNumberOfAllHalfEdges() < 2) {
					// found another error edge
					for (Polygon temp : e.getAdjacentPolygons()) {
						AtomicInteger count = errorCountMap.computeIfAbsent(temp, k -> new AtomicInteger(0));
						count.incrementAndGet();
					}
				}
			}
			index++;
		}
	}

	@Override
	public HealNonManifoldEdgeError createNew() {
		return new HealNonManifoldEdgeError();
	}

	@Override
	public HealingID getID() {
		return HealingID.S_NON_MANIFOLD_EDGE;
	}

}
